/** //<>//
 Classe que actua com a Controlador Principal, implementa els mètodes: settings, setup, draw i serialEvent .
 - Instancia tots els controladors que ens calen per a crear el programa.
 - Escolta la interacció de l'usuari amb la interficie i delega la feina a aquests controladors.
 - Esta equipada amb una base de dades,la classe BDPrincipal i un Model, la classe ModelPrincipal,
 per a gestionar el tractament i format dels registres.
 - Tambè té la classe Log que permet generar un informe de la linia de codi on s'executa.
 */
import java.lang.Exception.*; 
import java.util.Objects;
import java.lang.StackTraceElement;
import processing.serial.*;
import shapes3d.*;
import controlP5.*;

private Serial port;
private ControlP5 controlP5;
private ModelPrincipal modelPrincipal;
private ControladorPanel controladorPanel;
private ControladorGrafica controladorGrafica;
private Controlador3D controlador3D;
private ControladorGravetat controladorGravetat;
private ControladorCamera controladorCamera;

void settings() {
  //Primer, cal carregar les mides de la finestra i model de renderitzat.
  size (1150, 550, P3D);
  //Si la màquina pot, volem una definició alta. 
  smooth(8);
}

void setup() {
  //La base de dades desa les mides de disseny. Han de ser les mateixes, sinó, no fem rés.
  if (BDPrincipal.verificaMides(width, height)) {
    //Posició a la pantalla del PC de la finestra.
    surface.setLocation(BDPrincipal.x_pantalla, BDPrincipal.y_pantalla);

    //Instanciem els objectes que farem servir.
    controlP5 = new ControlP5(this);
    modelPrincipal = new ModelPrincipal();
    controladorGrafica = new ControladorGrafica(modelPrincipal);
    controladorPanel = new ControladorPanel(modelPrincipal);
    controlador3D = new Controlador3D(this, modelPrincipal);
    controladorGravetat = new ControladorGravetat(modelPrincipal);
    controladorCamera= new ControladorCamera();

    //Fem la premissa  que no tenim connexió.
    boolean portOk = false;
    try {
      //Iniciem al comunicació amb la placa. Atenció: si no rebem dades, 
      //cal comprovar si el codi de la placa té activat el flag que envia les dades a la targeta SD.
      port = new Serial(this, "COM6", 115200);
      //llegim fins (ASCII 10) abans de cridar a serialEvent.
      port.bufferUntil('\n');
      //si el port no està actiu, saltarà primer l'excepció
      portOk = port.active();
    } 
    catch (java.lang.RuntimeException e) { 
      //Si no tenim port, ho notifiquem a la consola.
      Log.printInfo(new Exception(), "port KO !!! " +e.toString());
    }
    //Demanem al controlador que inici la rutina.
    //Si el port no esta disponible, per defecte, carreguem  un arxiu.
    controladorGrafica.init(portOk);
  }
}

void draw() {
  //Color de fons
  background(255, 255, 255);
  //Pintem l'Escena 3D
  controlador3D.pintaEscena3D();
  //Pintem l'escena 2D
  controladorGrafica.pintaEscena2D();
}

/**
 Mètode per a gestionar la interacció quan movem el ratolí.
 */
void mouseMoved() {
  //Quan movem el ratolí, activem el flag de les zones per on passem.
  controladorGrafica.setZonaGrafica(
    (mouseX <= controladorGrafica.BD.left +controladorGrafica.BD.x_inici_dials + modelPrincipal.idRegistre*4)
    && (mouseX > controladorGrafica.BD.left + controladorGrafica.BD.x_inici_dials) 
    && mouseY > controladorGrafica.BD.top_zona && mouseY < controladorGrafica.BD.bottom_zona  );

  controlador3D.setZona3D((mouseX <controlador3D.BD.rigth)  
    && (mouseY >controlador3D.BD.top &&  mouseY <controlador3D.BD.bottom) );
  //Si estem en pause, pintem.
  if (controladorGrafica.isPause()) loop();
}

/**
 Mètode per a gestionar la interacció quan premem els botons del ratolí.
 */
void mousePressed() {
  // A la Zona 3D  Quan pressionem el ratolí  enviem la posició al controladorCamera.
  if (controlador3D.isZona3D()) 
    controladorCamera.capturaMouse(mouseX, mouseY);
  // Si la gràfica esta en pause, pintem la nova situació.
  if (controladorGrafica.isPause()) loop();
}

/**
 Mètode per a gestionar la interacció quan cliquem el botò dret.
 */
void mouseClicked() {
  //si estemen pause i cliquem en la zona de la grafica.
  if (controladorGrafica.isPause() &&  controladorGrafica.isZonaGrafica()) {
    // fem la rutina corresponent i tornem a pintar.
    controladorGrafica.getModel().seleccionarPunt(mouseY);
    loop();
  }
}

/**
 Mètode per a gestionar la interacció quan rodem el botò central.
 */
void mouseWheel(MouseEvent e) {
  //si estemen pause....
  if (controladorGrafica.isPause()) {
    //... i el fem rodar en la zona3D
    if (controlador3D.isZona3D()) 
      //desplacem l'identificador de l’últim  registre pintat.
      controladorGrafica.mouPunt(e.getCount());
    //pintem la nova situació. 
    loop();
  }
}

/**
 Mètode per a gestionar la interacció quan arrastrem el ratolí.
 */
void mouseDragged() {
  if (controlador3D.isZona3D()) {
    switch(mouseButton) {
    case LEFT:
      controladorCamera.rodarCamara(mouseX, mouseY);
      break;
    case CENTER:
      controladorCamera.mouObservador(pmouseX - mouseX);
      break;
    case RIGHT:
      controlador3D.desplasaVista3D(true);
      break;
    }
  }
  if (controladorGrafica.isPause() )loop();
}

/**
 Mètode per a gestionar què fem quan deixem anar el botò.
 */
void mouseReleased() {
  //Si estem a la zona 3D, capturem la posició de la camera.
  if (controlador3D.isZona3D())
    controladorCamera.desarAnglesCamera();
  //Desactivem el flag que permet moure l'escena 3D.
  controlador3D.desplasaVista3D(false);
}

/**
 Mètode per a llegir el buffer de dades del port Serial.
 */
void serialEvent (Serial port) { 
  if (port.available() > 0) {
    try {
      //llegim el registre i construïm una taula per cada valor separt per una coma.
      String registre[] = split(trim(port.readStringUntil('\n')), ',');
      if (registre != null)
        modelPrincipal.setRegistre(registre);
    } 
    //Capturem les excepcions
    catch (java.lang.ArrayIndexOutOfBoundsException e) {
      Log.printInfo(new Exception(), e.toString());
    }  
    catch (java.lang.NullPointerException e) {
      Log.printInfo(new Exception(), e.toString());
    }
  } else  Log.printInfo(new Exception(), "port.available()= " +port.available());
}
/**
 Mètode usat per l'objecte que implementa la interacció amb la llista desplagble
 que ens permet triar un arxiu.
 Atenció: el nom del mètode ha de ser el mateix que hem usat per a instanciar l'objecte.
 Si els noms no són iguals, el mètode no funciona.
 */
void llista_arxius(int idItem) {
  //Carreguem l'arxiu.
  controladorGrafica.obrirArxiu(idItem);
}

/**
 Mètode d'ajuda, usat per a dibuixar uns eixos en l'objecte passat per paràmetre
 i així observar en quina posició queden dibuixats en funció de l'orientació de l'objecte.
 Els colors sempre són Eix X = vermell, Eix Y = verd, Eix Z = blau.
 La zona negativa sempre es més curta.
 */
public static void pintaEixos(PGraphics g) {
  g.strokeWeight(3);
  g.stroke(255, 0, 0); 
  g.line(-10, 0, 0, 20, 0, 0);//Eix X
  g.stroke(0, 255, 0); 
  g.line(0, -10, 0, 0, 20, 0); //Eix Y
  g.stroke(0, 0, 255); 
  g.line(0, 0, -10, 0, 0, 20);// Eix Z
  g.strokeWeight(1);
}
/**
 Classe que fa de Base de Dades per a tots els objectes.
 */
public static final class BDPrincipal {
  //DIMENSIONS PANTALLA
  public static final int AMPLE = 1150;
  public static final int ALSADA = 550;
  //POSICIO DE LA PANTALLA
  public static final int x_pantalla = 45;
  public static final int y_pantalla = 30;
  /**
   Mètode per a verificar que les mides possades en el settings() són les que hi ha a la base de dades.
   */
  public static boolean  verificaMides(int x, int y ) {
    if ( x != Principal.BDPrincipal.AMPLE   ||  y != Principal.BDPrincipal.ALSADA ) {
      Log.printInfo(new Exception(), "ERROR!! les mides no coincideixen amb les que hi ha declarades a la BD");
      return false;
    }
    return true;
  }
}

/**
 Classe per a crear informació de la línia de codi on s’executa.
 */
public static final class Log {
  //Variable per a indexar l'enumeració de les files impreses.
  private static int ordre =0;
  /**
   * Mètode per a fer un log amb  un id per a saber:
   * l'ordre; nom de la classe;  nom del mètode; numero de linia i el missatge que volem veure impres.
   * Exemple d'ús :  Log.printInfo(new Exception(), [String amb el missatge que volem imprimir]);
   * @param e Exception
   * @param missatge per a imprimir.
   */
  public static void printInfo(Exception e, String missatge) {
    ordre ++;
    StackTraceElement llista = e.getStackTrace()[0];
    println(" id_ " + ordre, llista.getClassName() + " >> "+llista.getMethodName()+ ", fila:" + llista.getLineNumber()+ " > "+ missatge);
  }
}
